/*
 * lz.c
 *
   Copyright (C) 2002 Matthew T. Russotto
 * Copyright (C) 2012 Eric Biggers
 *
 * wimlib - Library for working with WIM files 
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#include "comp.h"



/* This function determines the best match for the string of characters
 * beginning at every index in the block/buffer.  The "best match" for a given
 * index is defined as the longest string of characters previously occurring in
 * the block that is the same as the string of characters starting at that
 * index.  If multiple matches tie for the longest length, the best one is taken
 * to be the last one (the one closest to the string being matched) to reduce
 * the number of bits needed to represent the match offset.
 *
 * match_offset_tab is an array indexed by buffer index.  For each character in
 * the buffer of offset i, match_offset_tab[i] will give the index of beginning
 * of the longest match for the string of characters beginning at that
 * character.
 *
 * For each character in the buffer of offset i, match_len_tab[i] will give
 * the length of beginning of the longest match for the string of
 * characters beginning at that character. 
 *
 * XXX This function takes up ~98% of the compression time. It should be
 * optimized somehow, perhaps by using a hash table like in zlib.
 */
void lz_analyze_block(const u8 block[], int block_len, 
				u16 match_offset_tab[], u16 match_len_tab[],
				unsigned max_match_len)
{
	/* Make a pass through the block to find matches of length 1.  This
	 * means that we determine, for each character in the block, the index
	 * at which that character had last previously appeared, relative to
	 * that character. */
	{
		/* chartab holds pointers to the last occurrence of each
		 * character. */
		const u8 *char_tab[256];
		ZERO_ARRAY(char_tab);

		for (int i = 0; i < block_len; i++) {
			u8 c = block[i];
			if (char_tab[c] != NULL) {
				/* Case where the character 'c' has already
				 * appeared in the block, so we can update the
				 * i-th entry of the match_offset_tab and the
				 * match_len_tab. */
				match_offset_tab[i] = &block[i] - char_tab[c];
				match_len_tab[i] = 1;
			} else {
				/* Case where the character 'c' has not appeared
				 * in the block yet. */
				match_len_tab[i] = 0;
			}
			char_tab[c] = &block[i];
		}
	}

	/* Now, make a pass through the block for each additional allowed match
	 * length, in order.  Every pass, we look for matches that are 1 longer
	 * than the matches we found previously.  The pass is done from the end
	 * of the block to the beginning, so that matches closer to the string
	 * being matched are always found before farther away matches.  */
	for (int match_len = 2; match_len <= max_match_len; match_len++) {

		/* Set to true if a match of length match_len is found.  If no
		 * such match is found, there is no need to search for longer
		 * matches, since there cannot be any. */
		bool found_a_match_of_len_match_len = false;

		for (int i = block_len - match_len; i >= 0; i--) {
			 /* For each character in the block (index i),
			  * match_offset_tab will contain the index of the best
			  * match for that location of no more than match_len -
			  * 1 bytes long, relative to that location.
			  * match_len_tab will contain the length of that match.
			  *
			  * If the length given in match_len_tab for index i is
			  * not equal to the previous match length (match_len -
			  * 1), then we have already found the longest match for
			  * the string beginning at index i, so searching for a
			  * longer one is not needed.
			  */

			if (match_len_tab[i] != match_len - 1)
				continue;

			/* c is the next character in the string that must be
			 * matched.  */
			u8 c = block[i + match_len - 1];

			/* Follow the chain of matches back to the beginning of
			 * the block to find a match for the string beginning
			 * at index i that matches match_len characters, all the
			 * way up to character c.  We can stop searching for a
			 * match whenever we get to a point when we know that
			 * there cannot be any previous match longer than
			 * match_len - 1 characters (when the match_len at the
			 * cursor index is less that match_len - 1). */
			u16 cursor_idx = i - match_offset_tab[i];
			while (match_len_tab[cursor_idx] == match_len - 1) {

				if (block[cursor_idx + match_len - 1] == c) {
					match_offset_tab[i] = i - cursor_idx;
					match_len_tab[i] = match_len;
					found_a_match_of_len_match_len = true;
					break;
				}
				cursor_idx -= match_offset_tab[cursor_idx];
			}
		}

		if (!found_a_match_of_len_match_len)
			break;
	}
}
